'use strict'

const url = require('url')
const createRouter = require('routes')
const concat = require('concat-stream')
const storage = require('./storage')
const bole = require('bole')
const logger = bole(process.title)('routes')
const hyperquest = require('hyperquest')
const json = require('./json')
const filenamify = require('filenamify')

module.exports = createServerRoutes

function respondWithError (code, res, err) {
  // TODO: proper error codes
  res.writeHead(code, { 'content-type': 'application/json' })

  const errorObject = {
    message: err.message
  }

  if (err.code) {
    errorObject.code = err.code
  }

  res.end(JSON.stringify(errorObject) + '\n')
}

function createServerRoutes (config, broadcast) {
  const router = createRouter()
  const storageUrl = config.storageUrl + config.storageAPIVersion

  router.addRoute('/heapsnapshot', function createHeapSnapshotRoute (req, res, match) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query

    if (method === 'head') {
      // TODO-2.x: res.writeHead(200, { content-length: <profile or snapshot length here!> })
      res.writeHead(200)
      res.end()
    } else if (method === 'get') {
      if (!params.asset || !params.app) { return }
      const readStream = storage.createAssetStream(config, { asset: params.asset })
      const filename = filenamify(params.app.replace(/\s+/g, '-') + '-' + params.asset + '.heapsnapshot')

      readStream.on('response', function downloadSnapshotOpen (storageRes) {
        if (storageRes.statusCode === 200) {
          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)
          readStream.pipe(res)
        } else {
          respondWithError(storageRes.statusCode, res, new Error('Error fetching asset'))
        }
      })

      readStream.on('error', function downloadSnapshotError (error) {
        respondWithError(404, res, error)
      })
    } else if (method === 'post') {
      if (!params.id) { return }
      storage.createSnapshot(config, { id: params.id }, function snapshotCreated (err, meta) {
        if (err) {
          return respondWithError(500, res, err)
        }

        res.setHeader('content-type', 'application/json')
        res.end(JSON.stringify(meta) + '\n')
      })
    } else {
      res.end()
    }
  })

  router.addRoute('/cpuprofile', function profileRoute (req, res, match, db) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query

    if (method === 'head') {
      // TODO-2.x: res.writeHead(200, { content-length: <profile or snapshot length here!> })
      res.writeHead(200)
      res.end()
    } else if (method === 'get') {
      if (!params.asset || !params.app) { return }
      const readStream = storage.createAssetStream(config, { asset: params.asset })
      const filename = filenamify(params.app.replace(/\s+/g, '-') + '-' + params.asset + '.cpuprofile')

      readStream.on('response', function downloadProfileOpen (storageRes) {
        if (storageRes.statusCode === 200) {
          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)
          readStream.pipe(res)
        } else {
          respondWithError(storageRes.statusCode, res, new Error('Error fetching asset'))
        }
      })

      readStream.on('error', function downloadProfileError (error) {
        respondWithError(404, res, error)
      })
    } else if (method === 'post') {
      if (!params.id) { return }

      let duration = 5
      if (params.duration) {
        const intermediary = parseInt(params.duration, 10)
        if (!isNaN(intermediary)) {
          duration = intermediary
        }
      }

      storage.createProfile(config, {
        id: params.id,
        duration: duration
      }, function profileCreated (err, meta) {
        if (err) {
          return respondWithError(500, res, err)
        }

        res.setHeader('content-type', 'application/json')
        res.end(JSON.stringify(meta) + '\n')
      })
    } else {
      res.end()
    }
  })

  // request/activate a license key
  router.addRoute('/licensekey', function updateLicenseKey (req, res, match) {
    const method = req.method.toLowerCase()

    if (method !== 'post') {
      return respondWithError(501, res, new Error('Method Not Implemented'))
    }

    req.pipe(concat((incomingPayload) => {
      let incoming
      try {
        incoming = JSON.parse(incomingPayload.toString())
      } catch (e) {
        return respondWithError(400, res, new Error('Invalid request'))
      }

      const missing = []
      if (isNullOrEmpty(incoming.firstName)) missing.push('firstName')
      if (isNullOrEmpty(incoming.lastName)) missing.push('lastName')
      if (isNullOrEmpty(incoming.email)) missing.push('email')
      if (isNullOrEmpty(incoming.company)) missing.push('company')
      if (isNullOrEmpty(incoming.notificationEmailDefault)) missing.push('notificationEmailDefault')
      if (isNullOrEmpty(incoming.marketingOptIn)) missing.push('marketingOptIn')

      if (missing.length !== 0) {
        const errorMessage = `missing property: ${missing.join(', ')}`
        return respondWithError(400, res, new Error(errorMessage))
      }

      const licenseInfo = {
        company: incoming.company,
        email: incoming.email,
        lastName: incoming.lastName,
        firstName: incoming.firstName,
        notificationEmailDefault: incoming.notificationEmailDefault,
        marketingOptIn: incoming.marketingOptIn,
        state: 'ACTIVE'
      }

      if (incoming.providerId) {
        licenseInfo.providerData = incoming.providerId
      }

      logger.debug(`storage.createLicense() => ${JSON.stringify(licenseInfo, null, 4)}`)
      const readStream = storage.createLicense(config, licenseInfo, function onError (error) {
        if (error) {
          respondWithError(400, res, new Error('Unable to create license'))
          return
        }
      })

      let statusCode
      let responded = false

      readStream.on('error', function (err) {
        if (responded) return
        responded = true
        respondWithError(400, res, err)
      })

      readStream.on('response', function (response) {
        statusCode = response.statusCode
        if (statusCode === 200) return

        if (responded) return
        responded = true
        respondWithError(400, res, new Error('status 400 creating license'))
      })

      readStream.pipe(concat((storageResponse) => {
        storageResponse = storageResponse.toString()
        logger.debug(`storage.createLicense() <= ${JSON.stringify(storageResponse, null, 4)}`)

        if (responded) return
        responded = true
        res.setHeader('content-type', 'application/json')
        res.end(storageResponse)
      }))
    }))
  })

  router.addRoute('/config/licensekey', function updateLicenseKey (req, res, match) {
    const method = req.method.toLowerCase()
    if (method === 'put') {
      req.pipe(concat(function gather (buffer) {
        hyperquest.put(storageUrl + '/license-info')
        .on('response', (response) => {
          response.pipe(res)
        })
        .on('error', (err) => {
          logger.error(err, 'Error processing /license-info')
          respondWithError(500, res, new Error('Internal error'))
        })
        .end(buffer.toString())
      }))
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/config/thresholds/:category/:id', function updateThreshold (req, res, match, db) {
    const method = req.method.toLowerCase()
    if (method === 'put') {
      req.pipe(concat(buf => {
        if (!match || !match.params || !match.params.id) { return }

        const thresholdConfig = json(buf.toString())
        logger.debug('received thresholds config', thresholdConfig)
        if (!thresholdConfig) {
          return respondWithError(400, res, new Error('Invalid JSON'))
        }

        // curl -X PUT -k -d '{"App Id Goes Here":{"cpu":{"enabled":true,"threshold":0,"delay":0,"delayUnits":"seconds","snooze":0,"snoozeUnits":"minutes","actions":{"notify":{"enabled":true,"method":"email","recipients":["max@nodesource.com"]},"snapshot":{"enabled":false},"profile":{"enabled":false,"duration":30}}}}}' https://localhost:4000/api/v1/thresholds
        const endpoint = storageUrl + '/thresholds'
        var payload = {}
        payload[match.params.id] = thresholdConfig

        const newThresholds = Object.assign({}, storage.getThresholds(), payload)

        hyperquest.put(endpoint)
        .on('error', (e) => {
          respondWithError(500, res, e)
        })
        .on('response', (response) => {
          res.writeHead(response.statusCode)
          res.end()
        })
        .end(JSON.stringify(newThresholds))
      }))
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/list', function listRoute (req, res, match) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query

    // allow callers to get back all hosts by passing in no id
    const query = !params.id ? { id: params.id } : {}

    if (method === 'get') {
      storage.getList(config, query).pipe(res)
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/settings', function settingsRoute (req, res, match) {
    const method = req.method.toLowerCase()
    if (method === 'get') {
      storage.getSettings(config).pipe(res)
    } else if (method === 'put') {
      req.pipe(concat(buf => {
        hyperquest.post(storageUrl + '/settings')
        .on('error', (e) => {
          respondWithError(500, res, e)
        })
        .on('response', (response) => {
          res.writeHead(response.statusCode)
          res.end()
        })
        .end(buf.toString())
      }))
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/license-info', function getLicenseInfo (req, res, match) {
    const method = req.method.toLowerCase()
    if (method === 'get') {
      storage.getLicenseInfo(config).pipe(res)
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/check-license', function checkLicense (req, res, match) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query
    if (method === 'get') {
      if (!params.license) { return }
      // curl -X POST -k https://localhost:4000/api/v1/check-license?license=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
      storage.checkLicense(config, { license: params.license }, function licenseChecked (error, payload) {
        if (error) {
          respondWithError(400, res, error)
        } else {
          res.write(payload)
          res.end()
        }
      })
    } else {
      respondWithError(501, res, new Error('Method Not Implemented'))
    }
  })

  router.addRoute('/versions', function versionsRoute (req, res) {
    const pkg = require('../../package.json')
    const versions = {
      console: pkg.version,
      storage: null,
      node: process.versions.node,
      nsolid: process.versions.nsolid
    }

    storage.getVersion(config, (err, version) => {
      // if couldn't get storage version, leave as null, and log
      if (err) {
        logger.error(err, 'Error getting storage version')
      } else {
        versions.storage = version
      }

      res.setHeader('content-type', 'application/json')
      res.setHeader('Access-Control-Allow-Origin', '*')
      res.end(JSON.stringify(versions))
    })
  })

  router.addRoute('/health', function checkLicense (req, res, match) {
    res.setHeader('content-type', 'application/json')
    res.setHeader('Access-Control-Allow-Origin', '*')
    res.end(JSON.stringify({ status: 'OK' }) + '\n')
  })

  return router
}

// Return whether a string is null or empty.
function isNullOrEmpty (string) {
  if (string == null) return true
  if (string === '') return true
  return false
}
