'use strict'

const hyperquest = require('hyperquest')
const through = require('through')
const url = require('url')
const path = require('path')
const concat = require('concat-stream')
const split = require('split')
const json = require('./json')
const bole = require('bole')
const logger = bole(process.title)('storage')
const unpipe = require('unpipe')

const QuantileHistory = require('./quantile_history')

const requestOpts = {
  method: 'GET',
  rejectUnauthorized: false
}

const reconnect = require('reconnect-core')(function (urlObj) {
  return hyperquest(url.format(urlObj), requestOpts)
})

module.exports.startSummaryData = startSummaryData
module.exports.getSummaryData = getSummaryData

module.exports.createAssetsStream = createAssetsStream
module.exports.createInfoStream = createInfoStream
module.exports.createMetricStream = createMetricStream
module.exports.createApplicationStream = createApplicationStream
module.exports.createApplicationsStream = createApplicationsStream
module.exports.createThresholdsStream = createThresholdsStream
module.exports.createVulnerabilitiesStream = createVulnerabilitiesStream

module.exports.createSnapshot = createSnapshot
module.exports.createProfile = createProfile
module.exports.createAssetStream = createAssetStream
module.exports.getList = getList
module.exports.getSettings = getSettings
module.exports.getLicenseInfo = getLicenseInfo
module.exports.checkLicense = checkLicense
module.exports.createLicense = createLicense
module.exports.getThresholds = getThresholds
module.exports.getVersion = getVersion

// cache of summary app metrics
const summaryAppMetrics = new Map()

function buildUrlObj (type, config, params) {
  const apiUrl = config.storageUrl + config.storageAPIVersion
  const urlObj = url.parse(apiUrl, true)
  urlObj.pathname = path.join(urlObj.pathname, type)
  urlObj.query = params || {}
  return urlObj
}

// Set this initially to true so an initial connected message is logged on
// startup, even if there was no problem
var errorsInPreviousCheck = true
var errorsSinceLastCheck = false
var lastData = ''
var connectionErrorTimer
function logIfConnected (config) {
  if (errorsSinceLastCheck) {
    var lastSuccess = ''
    if (lastData) {
      lastSuccess = ` Last success at ${new Date(lastData).toISOString()}`
    }
    logger.error(`Having trouble connecting to Storage at ${config.storageUrl}, trying again.${lastSuccess}`)
  } else if (errorsInPreviousCheck) {
    logger.info(`Connected to Storage at ${config.storageUrl}`)
  }
  errorsInPreviousCheck = errorsSinceLastCheck
  errorsSinceLastCheck = false
}
function createStorageStream (type, config, params) {
  const urlObj = buildUrlObj(type, config, params)
  logger.debug('creating storage stream', urlObj)

  if (!connectionErrorTimer) {
    connectionErrorTimer = setInterval(() => logIfConnected(config), 10000).unref()
  }

  const out = through((buf) => {
    const obj = json(buf.toString())

    if (!obj) {
      return
    }

    // if we reconnect to a graph stream, ensure we start with the
    // last known id.  If the browser client reconnects they will
    // setup all new subscriptions so that is already handled.
    if (obj.time) {
      urlObj.query.start = obj.time
    }

    out.queue(obj)
  })

  var prevStream, prevSplit
  const re = reconnect({
    immediate: true
  }, function (stream) {
    // split() returns a v1 stream so we have to use the npm unpipe
    if (prevSplit) unpipe(prevSplit)
    if (prevStream) prevStream.unpipe()
    const newSplit = stream.pipe(split())
    newSplit.pipe(out, {end: false})
    prevStream = stream
    prevSplit = newSplit
  })
  // .on('connect', () => logger.error('connected'))
  // .on('disconnect', () => logger.error('disconnected'))
  // .on('reconnect', () => logger.error('reconnected'))
  .on('error', (e) => {
    const errNumber = e.errno ? e.errno : 'UNDEFINED'
    logger.debug('(%s) failed to connect to storage server: %s, retrying', type, errNumber)
    errorsSinceLastCheck = true
  })

  re.connect(urlObj)

  // ensure destroying the stream works as expected
  const _destroy = out.destroy.bind(out)
  out.destroy = function avoidReconnect () {
    re.disconnect()

    _destroy()

    // TODO-2.x: this is not ideal, but hyperquest does not respect `.end` like
    // reconnect-core performs.  Another reason to bail on hyperquest
    re._connection.destroy()
  }

  // Attention: This switches the stream to flowing mode, however currently all
  // consumers of these streams are going to synchronously add their own 'data'
  // handler immediately after it is returned anyway.  Only two stream handlers
  // use pipe instead of data events, and those are consuming createAssetStream
  // which does not use this method.  If createStorageStream consumers are
  // changed to use pipe, this will need to be removed or changed to a
  // passthrough stream
  out.on('data', () => {
    // Since we're now seeing data wash out any recent errors and log if needed
    lastData = Date.now()
    if (errorsSinceLastCheck || errorsInPreviousCheck) {
      errorsSinceLastCheck = false
      logIfConnected(config)
    }
  })

  return out
}

function startSummaryData (config) {
  // maintains a history of quantiled cpu and memory use by application
  // this data is to populate the graphs on the Applications page
  const fields = ['cpu', 'heapUsed', 'heapSizeLimit']
  const metricStream = createStorageStream('metrics', config, {field: fields, start: '-150s'})
  metricStream.on('data', (record) => {
    let qh = summaryAppMetrics.get(record.app)
    if (qh == null) {
      qh = new QuantileHistory()
      summaryAppMetrics.set(record.app, qh)
    }
    qh.insert(record)
  })
}

function getSummaryData (app) {
  let qh = summaryAppMetrics.get(app)
  if (qh == null) {
    return {app: app, summary: []}
  }
  return {app: app, summary: qh.entries()}
}

function createMetricStream (config, params) {
  return createStorageStream('metrics', config, params)
}

function createApplicationsStream (config, params) {
  return createStorageStream('applications', config, params)
}

function createApplicationStream (config, params) {
  params = params || {}
  params.interval = '3s'
  return createStorageStream('application', config, params)
}

function createInfoStream (config, params) {
  return createStorageStream('info', config, params)
}

var CurrentThresholds = {}

function getThresholds () {
  return CurrentThresholds.thresholds || {}
}

function createThresholdsStream (config, params) {
  var storageStream = createStorageStream('thresholds', config, params)
  var pipedStream = storageStream.pipe(through(function (data) {
    CurrentThresholds = data
    this.queue(data)
  }))

  // Propagate destroy to the underlying stream
  var origDestroy = pipedStream.destroy.bind(pipedStream)
  pipedStream.destroy = () => {
    storageStream.destroy()
    origDestroy()
  }

  return pipedStream
}

function createVulnerabilitiesStream (config, params) {
  return createStorageStream('vulnerabilities', config, params)
}

// curl -k https://localhost:4000/api/v1/asset?asset=ebd2b348-8383-499a-b1c7-74c69114bd6b
function createAssetsStream (config, params) {
  const urlObj = buildUrlObj('assets', config, params)

  const out = through((buf) => {
    const obj = json(buf.toString())

    if (!obj) {
      return
    }

    out.queue(obj)
  })

  var hq = hyperquest.get(url.format(urlObj))
  hq.on('error', (e) => { logger.error(e, 'asset polling error') })
  .on('end', () => {
    // timeout = setTimeout(poll, 500)
  })
  .pipe(out, { end: false })

  function cleanup () {
    hq.destroy()
  }
  out.on('end', cleanup)
  out.on('close', cleanup)
  out.on('error', cleanup)

  // piped hyperquest will sometimes emit SSL errors during shutdown, so
  // silence its error logging
  const _destroy = out.destroy.bind(out)
  out.destroy = function () {
    hq.removeAllListeners('error')
    hq.on('error', () => {})
    hq.end()
    _destroy()
  }

  return out
}

function createSnapshot (config, params, fn) {
  const urlObj = buildUrlObj('snapshot', config, params)
  hyperquest.post(url.format(urlObj), (err, response) => {
    if (err) {
      return fn(err)
    }

    response.pipe(concat((payload) => {
      const obj = json(payload.toString())
      if (!obj) {
        return fn(new Error('Invalid response'))
      }

      fn(null, obj)
    }))
  })
  .on('error', fn)
  .end()
}

function createProfile (config, params, fn) {
  const urlObj = buildUrlObj('profile', config, params)
  hyperquest.post(url.format(urlObj), (err, response) => {
    if (err) {
      return fn(err)
    }

    response.pipe(concat((payload) => {
      const obj = json(payload.toString())
      if (!obj) {
        return fn(new Error('Invalid response'))
      }

      fn(null, obj)
    }))
  })
  .on('error', fn)
  .end()
}

function getList (config, params) {
  const urlObj = buildUrlObj('list', config, params)
  const req = hyperquest.get(url.format(urlObj))
  req.on('error', err => {
    logger.error('getList error', err)
  })
  return req
}

function getSettings (config, params) {
  const urlObj = buildUrlObj('settings', config, params)
  const req = hyperquest.get(url.format(urlObj))
  req.on('error', err => {
    logger.error('getSettings', err)
  })
  return req
}

function getLicenseInfo (config, params) {
  const urlObj = buildUrlObj('license-info', config, params)
  const req = hyperquest.get(url.format(urlObj))
  req.on('error', err => {
    logger.error('getLicenseInfo', err)
  })
  return req
}

function checkLicense (config, params, licenseChecked) {
  const urlObj = buildUrlObj('check-license', config, params)
  hyperquest.post(url.format(urlObj), (error, response) => {
    if (error) {
      licenseChecked(error)
    } else {
      response.pipe(concat((payload) => {
        licenseChecked(null, payload)
      }))
    }
  })
  .on('error', licenseChecked)
  .end()
}

function createLicense (config, params, fn) {
  const urlObj = buildUrlObj('license-info', config, params)
  const result = hyperquest.post(url.format(urlObj))
  result
    .on('error', fn)
    .end(JSON.stringify(params))
  return result
}

function createAssetStream (config, params) {
  const urlObj = buildUrlObj('asset', config, params)
  const req = hyperquest.get(url.format(urlObj))
  req.on('error', err => {
    logger.error('createAssetStream', err)
  })
  return req
}

// storage v1 version info endpoint
function getVersion (config, cb) {
  const onResponse = through((buf) => {
    const versionObject = json(buf.toString())
    if (versionObject == null) {
      // try v2 versions endpoint
      return getVersion2x(config, cb)
    }

    cb(null, versionObject.version)
  })

  hyperquest.get(config.storageUrl)
    .pipe(onResponse, { end: false })
    .on('error', cb)
}

// storage v2 version info endpoint
function getVersion2x (config, cb) {
  const onResponse = through((buf) => {
    const versionObject = json(buf.toString())
    if (versionObject == null) {
      // json function logs string prefix already
      return cb(new Error('unable to parse version info'))
    }

    if (versionObject.versions == null) {
      return cb(new Error(`invalid versions object ${JSON.stringify(versionObject)}`))
    }

    cb(null, versionObject.versions.program)
  })

  hyperquest.get(url.resolve(config.storageUrl, '/api'))
    .pipe(onResponse, { end: false })
    .on('error', cb)
}
