'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const uuid = require('uuid')
const EventEmitter = require('events')
const lodash = require('lodash')

exports.create = create

function create (context) {
  return new CommandDispatcher(context)
}

const cancelAttributes = ['id', 'requestId', 'command', 'filter']

class CommandDispatcher extends EventEmitter {
  constructor (context) {
    super()
    this._context = context
    this.defaultUnknownAgentTimeout = 1000
    this.defaultKnownAgentTimeout = 30000
  }

  handleMessage (message) {
    Logger.debug('Handling message', message)

    if (message.error) {
      const errmsg = message.error.message || 'No command response from agent.'
      this.emit(`error:${message.requestId}`, new Error(errmsg))
      return
    }

    this.emit(`data:${message.requestId}`, message)
    if (message.complete) {
      this.emit(`end:${message.requestId}`)
    }
  }

  /* Like send(message, timeout) below, but using a callback instead
   * of events.  The callback is optional.
   *
   * This method can only be used with messages that respond with a single
   * data response, so cannot be used with broadcast commands or commands that
   * send more than a single data response.
   */
  sendCB (message, timeout, cb) {
    const requestId = this.send(message, timeout)

    if (requestId == null) return null

    this.on(`data:${requestId}`, (data) => finishUp(null, data))
    this.on(`error:${requestId}`, (err) => finishUp(err))

    const self = this

    function finishUp (err, data) {
      self.removeAllListeners(`data:${requestId}`)
      self.removeAllListeners(`error:${requestId}`)

      if (cb == null) return

      err == null ? cb(null, data) : cb(err)
      cb = null
    }

    return requestId
  }

  /*
   * Send a command to an N|Solid Agent over ZeroMQ.
   * @params {object} message
   * @params {string} message.id - The ID of the agent.
   * @params {string} message.command - The command type.
   * @params {string} message.requestId - The request ID to use to associate
   * responses.
   * @params {object} [message.args] - Optional key-value pairs to send as
   * arguments to the command.
   * @params {object} [message.filter] - Optional key-value pairs for the
   * agent to filter on to determine whether or not to process the message.
   */
  send (message, timeout) {
    // process is shutting down.
    if (!this._context.zmqServer) return

    Logger.debug('Sending command message.', message)
    message = lodash.defaults({}, message, {
      requestId: uuid.v4()
    })

    var agent
    if (message.id) {
      agent = this._context.agentManager.byId(message.id)
      if (agent) {
        // If the agent is known and thought to be alive, choose a longer timeout
        timeout = timeout || this.defaultKnownAgentTimeout
      }
    }

    timeout = timeout || this.defaultUnknownAgentTimeout

    if (!message.command) {
      // can't have any listeners till this method returns the requestId,
      // so emit the error later
      setImmediate(() => {
        this.emit(`error:${message.requestId}`, new Error('Command required.'))
      })
      return message.requestId
    }

    var start = Date.now()
    Logger.debug(`dispatching ${message.command} to`, message.id || 'broadcast')
    this._context.zmqServer.send(message)

    // Open issue for broadcast commands, do not time those out.
    // https://github.com/nodesource/nsolid-storage/issues/131
    const timer = setTimeout(() => {
      const cancelMessage = lodash.pick(message, cancelAttributes)
      cancelMessage.cancel = true

      if (agent) {
        agent.pendingCommandTimers.delete(timer)
      }

      // This can be null during process shutdown.
      if (this._context.zmqServer) {
        this._context.zmqServer.send(cancelMessage)
      }

      this.emit(`error:${message.requestId}`, new Error('Command timeout.'))
      Logger.error(`Command timeout (${timeout} ms) for ${message.command} to ${message.id || 'broadcast'} with requestId ${message.requestId}.`)
    }, timeout)

    if (agent) {
      agent.pendingCommandTimers.add(timer)
    }

    this.on(`data:${message.requestId}`, () => {
      Logger.debug(`Command ${message.command} to ${message.id || 'broadcast'} with requestId ${message.requestId} finished after ${Date.now() - start} ms`)
      clearTimeout(timer)
      if (agent) {
        agent.pendingCommandTimers.delete(timer)
      }
    })
    this.on(`error:${message.requestId}`, () => {
      Logger.debug(`Command ${message.command} to ${message.id || 'broadcast'} with requestId ${message.requestId} errored after ${Date.now() - start} ms`)
      clearTimeout(timer)
      if (agent) {
        agent.pendingCommandTimers.delete(timer)
      }
    })

    return message.requestId
  }
}
