'use strict'

const fs = require('fs')
const path = require('path')
const childProcess = require('child_process')

const lodash = require('lodash')
const portFinder = require('portfinder')

const api = require('./api')
const util = require('../util')

exports.launch = launch

const Logger = util.getLogger(__dirname, __filename)

// Launch an instance of InfluxDB.
function launch (context, cb) {
  const config = context.config

  config.influxdb = {}
  config.influxdb.metaDir = path.join(config.data.dir, 'influxdb', 'meta')
  config.influxdb.dataDir = path.join(config.data.dir, 'influxdb', 'data')
  config.influxdb.walDir = path.join(config.data.dir, 'influxdb', 'wal')

  let err

  err = util.ensureDirExists(config.influxdb.metaDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.dataDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.walDir)
  if (err) return util.cbNextTick(cb, err)

  const opts = {
    host: 'localhost',
    port: 10001
  }
  portFinder.getPorts(2, opts, (err, ports) =>
    gotPorts(err, config, ports, cb)
  )
}

// Capture the ports needed for InfluxDB.
function gotPorts (err, config, ports, cb) {
  if (err) return cb(err)

  config.influxdb.portCluster = ports[0]
  config.influxdb.portHTTP = ports[1]

  configureInfluxDB(config, cb)
}

// Configure InfluxDB.
function configureInfluxDB (config, cb) {
  const logFile = config.logs.influx

  let logFD
  try {
    logFD = fs.openSync(logFile, 'a+')
  } catch (err) {
    Logger.error(err, `error opening InfluxDB log file: ${logFile}`)
    return cb(err)
  }

  Logger.info(`writing InfluxDB log to ${logFile}`)

  const env = {
    INFLUXDB_HOSTNAME: 'localhost',
    INFLUXDB_REPORTING_DISABLED: 'true',
    // following is apparently needed, but not used
    // See https://github.com/nodesource/nsolid-storage/issues/357
    INFLUXDB_BIND_ADDRESS: `localhost:${config.influxdb.portCluster}`,
    INFLUXDB_META_DIR: config.influxdb.metaDir,
    INFLUXDB_DATA_DIR: config.influxdb.dataDir,
    INFLUXDB_DATA_WAL_DIR: config.influxdb.walDir,
    INFLUXDB_ADMIN_ENABLED: false,
    INFLUXDB_HTTP_BIND_ADDRESS: `0.0.0.0:${config.influxdb.portHTTP}`,
    INFLUXDB_DATA_QUERY_LOG_ENABLED: false,
    INFLUXDB_HTTP_LOG_ENABLED: false
  }

  let proc

  try {
    proc = launchInfluxDB(env, logFD)
  } catch (err) {
    Logger.error(err, 'error launching InfluxDB ...')
    return cb(err)
  }

  const influxDBProcess = new InfluxDBProcess(proc, config)

  Logger.debug('waiting for InfluxDB to become ready ...')
  waitForInfluxDB(config, influxDBProcess, influxDBready)

  function influxDBready (err, version) {
    if (err) {
      Logger.error(`error waiting for InfluxDB to become ready: ${err}`)
      return cb(err)
    }

    function dbReady (err) {
      if (err) {
        Logger.error(err, 'error setting retention policy')
        return cb(err)
      }
      cb(null, influxDBProcess)
    }
    function dbCreated (err, body) {
      if (err) {
        Logger.error(err, 'error creating database nsolid')
        return cb(err)
      }
      createRp(config, dbReady)
    }
    createDb(config, dbCreated)
  }
}

function createDb (config, cb) {
  const createCommand = 'CREATE DATABASE nsolid'
  api.query(config.influxdb.portHTTP, 'nsolid', createCommand, cb)
}

function createRp (config, cb) {
  const getRetention = 'SHOW RETENTION POLICIES ON nsolid'
  api.query(config.influxdb.portHTTP, 'nsolid', getRetention, (err, body) => {
    if (err) {
      Logger.error(err, 'error checking existing retention policies')
      return cb(err)
    }

    // The result will look something like this:
    // { "results": [ {
    //   "series": [ {
    //     "columns": [ "name", "duration", "shardGroupDuration", "replicaN", "default" ],
    //     "values": [
    //       [ "default", "0", "168h0m0s", 1, false ],
    //       [ "agentData", "720h0m0s", "24h0m0s", 1, true ]
    //     ]
    //   } ]
    // } ] }

    var exists
    var retentionDays
    try {
      const response = JSON.parse(body)
      const policies = response.results[0].series[0].values
      const matching = policies.filter((pol) => { return pol[0] === 'agentData' })
      if (matching.length) {
        exists = true

        const match = matching[0][1].match(/^(\d+)h.*/)
        if (match) retentionDays = Math.round(parseInt(match[1], 10) / 24)
      }
    } catch (err) {
      Logger.error(err, 'error checking existing retention policies')
    }

    if (retentionDays != null) {
      Logger.info(`current retention policy: ${retentionDays} days`)
    }

    if (!exists || (retentionDays !== config.data.retentionDays)) {
      Logger.info(`setting retention policy: ${config.data.retentionDays} days`)
      const verb = exists ? 'ALTER' : 'CREATE'
      const retentionCommand =
        `${verb} RETENTION POLICY agentData ON nsolid DURATION ${config.data.retentionDays}d REPLICATION 1 DEFAULT`
      api.query(config.influxdb.portHTTP, 'nsolid', retentionCommand, cb)
    } else {
      cb()
    }
  })
}

function waitForInfluxDB (config, influxDBProcess, cb) {
  if (!influxDBProcess.isRunning()) {
    return cb(new Error('InfluxDB process terminated'))
  }

  if (config.influxdb.connectAttempt == null) {
    config.influxdb.connectAttempt = 0
  }

  config.influxdb.connectAttempt++
  if (config.influxdb.connectAttempt > 100) {
    return cb(new Error('unable to connect to InfluxDB'))
  }

  setTimeout(ping, 250)

  function ping () {
    Logger.debug('pinging InfluxDB to see if it\'s ready')
    api.ping(config.influxdb.portHTTP, onPing)
  }

  function onPing (err, version) {
    if (err) return waitForInfluxDB(config, influxDBProcess, cb)

    Logger.info(`InfluxDB version ${version} ready`)
    cb(null, version)
  }
}

// Launch the InfluxDB process.
function launchInfluxDB (env, logFD) {
  const platArch = `${process.platform}-${process.arch}`
  const influxPath = path.resolve(path.join(__dirname, '..', '..', 'tsdb', platArch))
  const influxExe = path.join(influxPath, 'influxd')

  env = lodash.defaults({}, env, process.env)

  const opts = {
    env: env,
    stdio: ['ignore', 'ignore', logFD]
  }

  return childProcess.spawn(influxExe, opts)
}

// This class manages the InfluxDB process.
class InfluxDBProcess {
  constructor (process, config) {
    this.process = process
    this.config = config
    this.port = config.influxdb.portHTTP
    this.url = `http://0.0.0.0:${this.port}`

    this.process.on('error', err => this._onError(err))
    this.process.on('exit', (code, signal) => this._onExit(code, signal))
  }

  isRunning () {
    return this.process != null
  }

  kill (cb) {
    if (!this.isRunning()) return

    if (cb) this.process.on('exit', cb)
    this.process.kill()
  }

  _onError (err) {
    Logger.error(err)
    this.process = null
  }

  _onExit (code, signal) {
    Logger.info(`InfluxDB exit(${code}) signal: ${signal}`)
    this.process = null
  }
}
