'use strict'

const async = require('async')
const url = require('url')
const path = require('path')
const doSnapshot = require('./actions/snapshot')
const doProfile = require('./actions/profile')
const doNotify = require('./actions/notify')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

class ThresholdAction {
  constructor (metric, value, cache, state, thresholdConfig, context, serverUrl) {
    this.metric = metric
    this.state = state
    this.cache = cache
    this.thresholdConfig = thresholdConfig
    this.context = context || cache.context
    this.serverUrl = serverUrl

    const consoleURL = this.context.config.notification.consoleURL

    this.actions = {
      snapshot: function performSnapshot (context, appConfig, actionConfig, proc, done) {
        Logger.debug('snapshot on process %s (%s)', proc.id, proc.app)
        appConfig.snapshotEndpoint = { url: serverUrl }
        doSnapshot(appConfig, proc.id, (e, r) => {
          if (e) {
            return done(e)
          }

          const heapUrl = url.resolve(consoleURL, '/' + path.join(
            'app',
            r.app,
            'heap',
            r.id,
            r.asset + ''
          ))

          done(null, heapUrl, r)
        })
      },
      profile: function performProfile (context, appConfig, actionConfig, proc, done) {
        const duration = actionConfig.profileDuration
        Logger.debug('profile on process %s (%s) for %s', proc.id, proc.app, duration)
        appConfig.profileEndpoint = { url: serverUrl }
        doProfile(appConfig, proc.id, duration, (e, r) => {
          if (e) {
            return done(e)
          }

          const profileUrl = url.resolve(consoleURL, '/' + path.join(
            'app',
            r.app,
            'profile',
            r.id,
            r.asset + '',
            'treemap'
          ))

          done(null, profileUrl, r)
        })
      },
      notify: function performNotify (results, appConfig, actionConfig, proc, done) {
        if (context.provider.isOptedOut()) {
          Logger.debug('Skipping notify due to opt-out.')
          return done()
        }

        const payload = {
          type: 'threshold',
          method: 'email',
          data: {
            metric: metric,
            value: value,
            threshold: thresholdConfig.threshold,
            recipients: actionConfig.recipients,
            delay: thresholdConfig.delay,
            delayUnits: thresholdConfig.delayUnits,
            process: proc,
            artifacts: results
          }
        }

        Logger.debug('notify on process %s (%s)', proc.id, proc.app)
        appConfig.notificationEndpoint = {
          url: context.isStaging
            ? 'https://staging-services.nodesource.io/notification'
            : 'https://services.nodesource.io/notification'
        }

        appConfig.license = {
          key: cache.config.license.key
        }

        doNotify(appConfig, payload, done)
      }
    }
  }

  getActions () {
    if (!this.thresholdConfig) {
      return false
    }

    const actions = this.thresholdConfig.actions
    if (!actions) {
      return false
    }

    return actions
  }

  execute (config, next) {
    const configuredActions = this.getActions()
    const actions = this.actions
    const metric = this.metric
    const cache = this.cache
    const state = this.state
    const context = this.context

    const processid = state.id

    if (!configuredActions) {
      return finish()
    }

    let willNotify = false

    const actionNames = Object.keys(configuredActions).filter((key) => {
      if (key === 'notify') {
        willNotify = !!configuredActions.notify.enabled
        return false
      }

      return true
    })

    async.mapSeries(actionNames, processAction, mapComplete)

    function processAction (actionName, done) {
      const action = actions[actionName]
      const actionConfig = configuredActions[actionName]

      if (!action || !actionConfig || !actionConfig.enabled) {
        return done()
      }

      action(context, config, actionConfig, state, (e, resourceUrl, result) => {
        if (e) {
          return done(null, {
            error: {
              message: e.message,
              code: e.code
            }
          })
        }

        if (!result && !resourceUrl) {
          return done()
        }

        done(null, {
          type: actionName,
          result: result,
          url: resourceUrl
        })
      })
    }

    function notify (results, fn) {
      actions.notify(results, config, configuredActions.notify, state, (e, r) => {
        const artifact = {
          type: 'notify'
        }

        if (e) {
          artifact.error = {
            message: e.message,
            code: e.code
          }
        } else {
          artifact.result = r
        }

        results.push(artifact)

        fn(e, results)
      })
    }

    function mapComplete (err, results) {
      if (willNotify && results) {
        return notify(results.filter(Boolean), finish)
      }

      finish(err, results)
    }

    function finish (e, r) {
      const proc = cache.getProcess(processid)

      // if the process crashed, avoid exploding
      if (!proc) {
        return next()
      }

      if (!proc.thresholds) {
        proc.thresholds = {}
      }
      proc.thresholds[metric] = { snoozeStart: Date.now() }
      next(e, r)
    }
  }
}

module.exports = ThresholdAction
