'use strict'

const Buffer = require('buffer').Buffer
const EventEmitter = require('events')
const util = require('../util')

const Logger = util.getLogger(__dirname, __filename)

// A first-in, first-out buffer queue that drops the oldest messages when full.

class BufferQueue extends EventEmitter {
  constructor (params) {
    super()

    this.name = params.name
    this.config = params.config
    this.sendBuffered = params.sendBuffered

    // Intialize these on the object in the constructor, but set via setOptions
    this.enabled = true
    this.logSizeInterval = 0
    this.setOptions(this.config)

    this.bufferSize = 0 // In Bytes
    this.buffering = true

    // An oldest -> newest one-way linked list with a shortcut to append
    // at the end (last)
    this.linkedlist = {
      // further entries on the `next` chain are newer, this is the oldest
      next: undefined,
      last: undefined,
      buffers: undefined,
      length: 0
    }

    // bufferSize interval logger, using captured scope for convinence
    this.logBufferSize = () => {
      const bufSizeMb = this.bufferSizeMb()
      // Only log if size is "significant".
      if (bufSizeMb >= 0.001) {
        Logger.info(`${bufSizeMb}mb buffered for socket ${this.name}`)
      }
    }

    // Start interval logger if configured.
    if (this.logSizeInterval) {
      if (this.loggingTimer) clearInterval(this.loggingTimer)
      this.loggingTimer = setInterval(this.logBufferSize, this.logSizeInterval)
      this.loggingTimer.unref()
    }
  }

  pause () {
    if (this.buffering) return
    this.buffering = true
    Logger.info(`${this.name} paused. Buffering enabled. (${this.bufferSizeMb()}mb)`)

    // Restart interval logger if configured.
    if (this.logSizeInterval) {
      if (this.loggingTimer) clearInterval(this.loggingTimer)
      this.loggingTimer = setInterval(this.logBufferSize, this.logSizeInterval)
      this.loggingTimer.unref()
    }
  }

  resume () {
    // Cancel logging timer if connected.
    clearInterval(this.loggingTimer)
    Logger.info(`${this.name} resumed. Flushing buffered messages. (${this.bufferSizeMb()}mb)`)
    // Flush messages over the wire recursing in Immediate (if there are any)
    // This will set buffering to "off" once all messages are flushed.
    setImmediate(() => {
      this.flushNextBufferAsync(this.linkedlist.next)
      this.linkedlist.next = undefined
    })
  }

  setOptions (config) {
    // -1 means automatically determined, other negative values are not recognized as valid
    this.enabled = config.maxBufferSizeMb > 0 || config.maxBufferSizeMb === -1

    this.logSizeInterval = config.logSizeInterval * 1000 // Convert to ms
    if (this.logSizeInterval < 0 || Number.isNaN(this.logSizeInterval)) {
      this.logSizeInterval = 0
    }
  }

  // Clean up
  close () {
    clearInterval(this.loggingTimer)
  }

  // Add buffers to the queue.
  queueBuffers (buffers) {
    if (!this.enabled) {
      // Queueing disabled
      return
    }

    let buffersSize = 0
    // Sum up the size of the incoming buffers.
    for (let i = 0; i < buffers.length; i++) {
      buffersSize += Buffer.byteLength(buffers[i], 'utf8')
    }

    // Correct the bufferSize to the new amount
    this.bufferSize += buffersSize

    this.emit('append', buffersSize)

    // Construct a new item for list insertion
    const bufferItem = {
      next: undefined,
      buffers,
      buffersSize
    }
    if (this.linkedlist.last) {
      // List exists, append.
      this.linkedlist.last.next = bufferItem
      this.linkedlist.last = bufferItem
    } else {
      // Begin the list.
      this.linkedlist.next = bufferItem
      this.linkedlist.last = bufferItem
    }
    this.linkedlist.length += 1
  }

  // Clean out messages starting at the oldest in a predefined batch size.
  cleanupLinkedList (goalSize) {
    if (!goalSize) goalSize = this.cleanupSize * 1000 * 1000

    Logger.info(`Cleaning out linkedlist on ${this.name} with a cleanupSize of ${this.cleanupSize} on ${this.name}`)

    // Testing purposes only.
    this.emit('pre-cleanup')

    Logger.info(`ll size on ${this.name}, pre-cleanup: ${this.linkedlist.length}`)

    let progress = 0
    let i = 0
    // If there isn't a next the end of the linkedlist was reached.
    while (progress < goalSize && this.linkedlist.next) {
      const current = this.linkedlist.next

      // Correct the bufferSize for the buffers that are being removed
      this.subtractBufferSize(current.buffersSize)
      progress += current.buffersSize

      // linkedlist remove oldest
      const next = current.next
      this.linkedlist.next = next

      i++
    }
    const progressMb = (progress / 1000 / 1000).toFixed(3)
    Logger.info(`Cleaned out ${i} items (${progressMb}mb) from linkedlist on ${this.name}`)

    return { count: i, progressBytes: progress }
  }

  cleanupLinkedListLength (goalCount) {
    if (!goalCount) goalCount = 10000

    Logger.info(`Cleaning out linkedlist on ${this.name} due to VM limits`)

    // Testing purposes only.
    this.emit('pre-cleanup')

    let progress = 0
    let i = 0
    // If there isn't a next the end of the linkedlist was reached.
    while (i < goalCount && this.linkedlist.next) {
      const current = this.linkedlist.next

      // Correct the bufferSize for the buffers that are being removed
      this.subtractBufferSize(current.buffersSize)
      progress += current.buffersSize

      // linkedlist remove oldest
      const next = current.next
      this.linkedlist.next = next
      i++
    }
    const progressMb = (progress / 1000 / 1000).toFixed(3)
    Logger.info(`Cleaned out ${i} items (${progressMb}mb) from linkedlist on ${this.name}`)

    return { count: i, progressBytes: progress }
  }

  // flush out this buffer and schedule next for next Immediate
  flushNextBufferAsync (bufferItem) {
    // If the item doesn't exist that means the linkedlist did not have
    // a starting item, log & bail.
    if (!bufferItem) {
      this.buffering = false
      Logger.info(`${this.name} socket had no buffered messages. (${this.bufferSizeMb()}mb)`)
      return
    }

    this.emit('flush', bufferItem.buffersSize)

    // Correct the bufferSize for the buffers that are being flushed
    this.subtractBufferSize(bufferItem.buffersSize)

    // Send buffers over the wire
    this.sendBuffered(bufferItem.buffers)

    // If the item doesn't have a valid `.next`, we are at the end. Log & bail.
    if (!bufferItem.next) {
      if (this.linkedlist.last !== bufferItem) {
        Logger.warn(`${this.name} socket linkedlist did not flush correctly!`)
      }
      this.linkedlist.last = undefined
      // Set buffering to "off"
      this.buffering = false
      Logger.info(`${this.name} socket finished flushing all buffered messages. (${this.bufferSizeMb()}mb)`)
      return
    }

    // Async recurse with the next item after I/O
    setImmediate(() => {
      this.flushNextBufferAsync(bufferItem.next)
      bufferItem.next = undefined
    })
  }

  // Subtract a number of bytes from the bufferSize
  // and warn if it drops below 0
  subtractBufferSize (sub) {
    this.bufferSize -= sub
    if (this.bufferSize < 0) {
      Logger.warn(`${this.name} emptied more megabytes than it buffered!`)
      this.bufferSize = 0
    }

    this.linkedlist.length -= 1
    if (this.linkedlist.length < 0) {
      Logger.warn(`${this.name} emptied more list items it buffered!`)
      this.linkedlist.length = 0
    }
  }

  // Get the bufferSize in Megabytes, truncate decimals only to Kilobytes
  bufferSizeMb () {
    return (this.bufferSize / 1000/* kb */ / 1000/* mb */).toFixed(3)
  }
}

module.exports = BufferQueue
