'use strict'

const util = require('../util')
const ZeroMQServer = require('./server')
const Logger = util.getLogger(__dirname, __filename)

exports.launch = launch

function launch (context, cb) {
  const config = context.config
  const handlers = require('./handlers')(context)
  Logger.info(`ZeroMQ data channel url: ${config.sockets.dataBindAddr}`)
  Logger.info(`ZeroMQ bulk channel url: ${config.sockets.bulkBindAddr}`)
  Logger.info(`ZeroMQ command channel url: ${config.sockets.commandBindAddr}`)

  const server = new ZeroMQServer({
    config: config,
    preBindCb: (name, socket) => {
      addHandlers(name, socket, handlers, context)
    },
    bindCb: err => {
      if (err) return cb(err)
      cb(null, server)
    }
  })
}

function addHandlers (name, socket, handlers, context) {
  if (name === 'command') {
    socket.on('message', message => {
      // An XPUB message is a subscription.  The first byte in the buffer will be
      // 1 for subscribe and 0 for unsubscribe.  The remaining bytes are the
      // subscription string.
      var isSub = message[0]
      var subscription = message.slice(1).toString()
      if (isSub && /^configuration/.test(subscription)) {
        socket.send([subscription, JSON.stringify({ config: context.config })])
      }
    })
  }

  if (name === 'data') {
    socket.on('message', message => {
      try {
        message = JSON.parse(message.toString('utf8'))
      } catch (err) {
        return Logger.error('ZeroMQ data message parse error', err, message.toString())
      }

      fixTimeProperties(message)

      // Each handler has a predicate which looks at attributes on the
      // message and determines whether or not it should run.
      handlers.dynamic.forEach(handler => {
        let shouldRun

        try {
          shouldRun = handler.predicate(message)
        } catch (err) {
          return Logger.error('ZeroMQ message handler predicate error', err, message)
        }

        if (shouldRun) {
          setImmediate(() => {
            try {
              handler.run(message)
            } catch (err) {
              Logger.error('ZeroMQ message handler error', err)
            }
          })
        }
      })
    })
  }

  if (name === 'bulk') {
    socket.on('message', (envelope, data) => {
      try {
        envelope = JSON.parse(envelope)
      } catch (err) {
        return Logger.error('ZeroMQ bulk envelope parse error', err)
      }

      handlers.bulk.run(envelope, data)
    })
  }
}

// Convert agents `time: {seconds: nanoseconds:}` into usable things.
// Note both `time` properties are strings, and nanoseconds is left-zero-padded.
function fixTimeProperties (object) {
  // leave early unless a number of constraints hold
  if (object.recorded == null) return

  const recorded = object.recorded
  if (recorded.seconds == null) return
  if (recorded.nanoseconds == null) return

  // get the numeric values of the properties
  const seconds = parseInt(recorded.seconds, 10)
  const nanoseconds = parseInt(recorded.nanoseconds, 10)

  // final chance to give up
  if (!Number.isInteger(seconds) || !Number.isInteger(nanoseconds)) return

  object.time = Math.round(seconds * 1000 + nanoseconds / 1000000)
  object.timeNS = `${recorded.seconds}${recorded.nanoseconds}`
}
