'use strict'

const util = require('../util')
const zmq = util.requireZmq()

const nsolidStorage = require('../../nsolid-storage')

const Logger = util.getLogger(__dirname, __filename)

class ZMQBindSocket {
  constructor (params) {
    this.bindUrl = params.bindUrl
    this.name = params.name
    this.type = params.type
    this.bindCb = params.bindCb
    this.preBindCb = params.preBindCb

    if (!/:\/\//.test(this.bindUrl)) {
      this.bindUrl = 'tcp://' + this.bindUrl
    }

    if (/^(sub|push)$/.test(this.type)) {
      Logger.error(`Use ZMQRemoteSocket for ${this.type} sockets (${params.bindUrl || params.remoteUrl})`)
      nsolidStorage.shutdown(1)
    }

    this.socket = zmq.socket(params.type)

    this.socket.monitor(500, 0)
    this.socket.setsockopt(zmq.ZMQ_SNDHWM, params.hwm)
    this.socket.setsockopt(zmq.ZMQ_IPV6, 1)

    // Make every subscription fire a message even if it's redundant.  Without
    // this, agents trying to get configuration at the same time will fail.
    if (this.type === 'xpub') {
      this.socket.setsockopt(zmq.ZMQ_XPUB_VERBOSE, 1)
    }

    this.socket.curve_server = 1
    this.socket.curve_secretkey = params.privateKey

    this.socket.on('bind', () => {
      Logger.debug(`${this.type} socket bound to ${this.bindUrl}`)
      if (this.bindCb) {
        this.bindCb()
      }
    })
    this.socket.on('bind_error', (event, endpoint, ex) => this.exitOnError(ex || event))

    if (this.preBindCb) {
      this.preBindCb(this.name, this.socket)
    }
    this.socket.bind(this.bindUrl)

    this.socket.on('accept', () => Logger.debug(`${this.type} socket accepted a remote connection to ${this.bindUrl}`))
  }

  exitOnError (err) {
    if (!(err instanceof Error)) err = new Error(`zmq bind error: ${err}`)
    Logger.error(err, `Couldn't bind ${this.type} channel socket to ${this.bindUrl}`)
    if (this.bindCb) {
      this.bindCb(err)
    } else {
      nsolidStorage.shutdown(1)
    }
  }

  close () {
    // Already closed
    if (!this.socket) return

    this.socket.on('unbind', () => {
      this.socket.close()
    })
    this.socket.on('close', () => {
      this.socket.removeAllListeners()
      this.socket.unmonitor()
      this.socket = null
    })
    this.socket.setsockopt(zmq.ZMQ_LINGER, 0)
    this.socket.unbind(this.bindUrl)
  }
}

module.exports = ZMQBindSocket
