'use strict'

const Client = require('nsolid-apiclient')
const commands = require('nsolid-apiclient/commands')
const debug = require('util').debuglog('cli')
const getLimit = require('./get-limit')
const countingStream = require('./counting-stream')

class CliCommand {
  constructor (command, remoteSocket, opts) {
    this._client = new Client(remoteSocket)
    this._command = command
    this._opts = opts
    this._lines = 0
    this._limit = Infinity
    this._request = undefined
    this._stream = undefined

    this._normalizeTimeout()
    this._determineLimitSettings()
  }

  execute () {
    if (this._shouldWrite()) return this._write()
    if (this._shouldExec()) return this._tryExec()
    return this._read()
  }

  _shouldWrite () {
    return (this._opts.body != null) && commands.writeCommands.has(this._command)
  }

  _shouldExec () {
    return this._opts.force || !commands.readCommands.has(this._command)
  }

  _tryExec () {
    if (commands.execCommands.has(this._command)) return this._exec()
    const forceString = (this._opts.force) ? ' using --force' : ''
    throw new Error(`Unable to run command '${this._command}'${forceString}`)
  }

  _write () {
    this._request = this._client.write(this._command, this._opts)
    this._stream = this._request
    return this._stream
  }

  _read () {
    this._request = this._client.read(this._command, this._opts)
    this._stream = this._request
    return this._stream
  }

  _exec () {
    this._request = this._client.exec(this._command, this._opts)
    // If user specified --no-asset then just return the command-result
    if (commands.assetCommands.has(this._command) && this._opts.asset !== false) {
      this._request.fetchAsset = true
    }

    // by default responses are streaming and don't terminate
    if (!this._applyLimit) {
      this._stream = this._request
      return this._stream
    }

    // initiate new line counter to determine when line limit is hit
    if (this._applyLimit) {
      const counting = countingStream()
      this._stream = this._request.pipe(counting)
      this._request.on('response', res => {
        this._stream.emit('response', res)
      })
      counting.on('lines', lines => this._onlines(lines))

      // In case of --limit=auto, we try to guess how many lines the limit should be
      if (this._autolimit) {
        getLimit(this._client, this._opts, (err, limit) => this._onlimit(err, limit))
      }
    }

    return this._stream
  }

  _onlines (lines) {
    this._lines = lines
    this._checkDone()
  }

  _onlimit (err, limit) {
    if (err) return this._stream.emit('error', err)
    this._limit = limit
    this._checkDone()
  }

  _checkDone () {
    debug({ lines: this._lines, limit: this._limit })
    if (this._lines >= this._limit) this._request.response.destroy()
  }

  _determineLimitSettings () {
    if (this._opts.limit === 'auto') {
      // auto limit set via `--limit=auto`
      this._autolimit = true
      this._limit = Infinity
      this._applyLimit = true
    } else if (typeof this._opts.limit === 'number') {
      // manual limit set via `--limit=x`
      this._autolimit = false
      this._limit = this._opts.limit
      if (this._limit < 1) {
        throw new Error(
          `Invalid limit ${this._limit}. Limit needs to be 'auto' or a number greater than zero.`
        )
      }
      this._applyLimit = true
    } else {
      // no limit set, stay in streaming mode
      this._autolimit = false
      this._limit = Infinity
      this._applyLimit = false
    }
    debug({ applyLimit: this._applyLimit, autolimit: this._autolimit, limit: this._limit })
  }

  _normalizeTimeout () {
    if (this._opts.timeout) {
      // convert to millis
      this._opts.timeout = this._opts.timeout * 1000
      debug('setting timeout to %sms', this._opts.timeout)
    }
  }
}

module.exports = function cli (command, remoteSocket, opts) {
  return new CliCommand(command, remoteSocket, opts).execute()
}
